"""nova_utils.data.handler

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06th September 2023

The nova-utils data handler package implements different handler to load and save data to various backends.
"""